#!/usr/bin/perl
# rp.pl
# Perl script to parse rp.log files from XP Restore Points
#
# In order to use this script effectively, you need a means by which
# to mount the acquired image as a file system (ie, SmartMount, ImDisk, 
# etc.) and the SysInternals psexec.exe tool.  Mount the acquired image to
# a read-only drive letter (ie, G:\) and then open a command prompt and
# type "psexec -s cmd"...this will open the command prompt to the 
# C:\Windows\system32 directory, with the process running w/ System privs.
# 
# In this System-level command prompt, 'cd' to the G:\ drive and type 'cd 
# sys*'...this will change directories to the System Volume Information dir
# (unless there's another directory that starts with 'sys').  Then type 'cd 
# _restor*'...this will change directories to the _restore{<GUID>} directory.
# Highlight and copy this entire path to the clipboard.  Now, 'cd' to the 
# directory where your copy of rp.pl is located and enter 'rp.pl -d', and the
# paste the path IN QUOTES into the command line.  Add your options as 
# appropriate.
# 
#
# copyright 2009 H. Carvey, keydet89@yahoo.com
#-----------------------------------------------------------

use strict;
use Getopt::Long;

my %config;
Getopt::Long::Configure("prefix_pattern=(-|\/)");
GetOptions(\%config,qw(dir|d=s csv|c tln|t sys|s=s user|u=s help|?|h));

if ($config{help} || !%config) {
	usage();
	exit 1;
}

die "You must include the path to the RP* directories.\n" unless ($config{dir});

my $rpdir = $config{dir};
die $rpdir." not found.\n" unless (-e $rpdir && -d $rpdir);
$rpdir = $rpdir."/" unless ($rpdir =~ m/\/$/);

my %rps;
opendir(DIR,$rpdir);
while (my $dir = readdir(DIR)) {
	next unless ($dir =~ m/^RP/ && -d $rpdir.$dir);
	my $tag = $dir;
	$tag =~ s/RP//;
	$rps{$tag} = $rpdir.$dir;
}
closedir(DIR);

foreach my $r (sort keys %rps) {
	my $file = $rps{$r}."/rp\.log";
	if (-e $file) {
		my ($time,$type,$str) = getRpData($file);
		
		if ($config{csv}) {
			print "RP".$r.",".gmtime($time)." Z,".$type.",".$str."\n";
		}
		elsif ($config{tln}) {
			print $time."|RP|".$config{sys}."|".$config{user}."|Restore Point created - ".$str."\n";
		}
		else {		
			print "RP".$r."  ".gmtime($time)." Z  ".$str."\n";
		}
	}
}

#-------------------------------------------------------------
# getTime()
# Translate FILETIME object (2 DWORDS) to Unix time, to be passed
# to gmtime() or localtime()
#-------------------------------------------------------------
sub getTime($$) {
	my $lo = shift;
	my $hi = shift;
	my $t;

	if ($lo == 0 && $hi == 0) {
		$t = 0;
	} else {
		$lo -= 0xd53e8000;
		$hi -= 0x019db1de;
		$t = int($hi*429.4967296 + $lo/1e7);
	};
	$t = 0 if ($t < 0);
	return $t;
}

#-------------------------------------------------------------
# getRpData()
#
#-------------------------------------------------------------
sub getRpData {
	my $file = $_[0];
	my $data;
	
	my %types = (0 => "Application Install",
		        1 => "Application Uninstall",
		        7 => "System CheckPoint",
		        10 => "Device Driver Install",
		        11 => "System Checkpoint",
		        12 => "Modify Settings",
		        13 => "Cancelled Operation");
		        
	my ($rp_time,$rp_type,$rp_str);
	
	eval {
		open(FH,"<",$file);
		binmode(FH);
		seek(FH,0x210,0);
		read(FH,$data,8);
		my @t = unpack("VV",$data);
		$rp_time = getTime($t[0],$t[1]);
				
		seek(FH,0x04,0);
		read(FH,$data,4);
		$rp_type = unpack("V",$data);
	
		my $offset = 0x10;
		my $tag = 1;
		my @strs;
		while($tag) {
			seek(FH,$offset,0);
			read(FH,$data,2);
			if (unpack("v",$data) == 0) {
				$tag = 0;
			}
			else {
				push(@strs,$data);
			}
			$offset += 2;
		}
		$rp_str = join('',@strs);
		$rp_str =~ s/\00//g;

	};
	return ($rp_time,$types{$rp_type},$rp_str);
}

#---------------------------------------------------------
# usage()
#---------------------------------------------------------
sub usage {
	print<< "EOT";
RP [options]
Parse Windows XP Restore Point rp.log files in binary mode, translating the
information available; sends data to STDOUT, can also generate timeline data

  -d directory.............path to directory where RP* subdirs are found
  -c ......................output in csv format to STDOUT
  -t ......................output in timeline format to STDOUT (redirect to an
                           event file)
  -s system................add system name to appropriate field in tln 
                           output
  -u user..................add user name (or SID) to appropriate field in 
                           tln output
  -h.......................Help (print this information)
  
Ex: C:\\>rp -d "G:\\System Volume Information\\_restore{<GUID>}" 
    C:\\>rp -d D:\\case\\restore -t 
  
copyright 2009 H. Carvey
EOT
}
